use axum::{
    body::Body,
    http::{self, StatusCode, header},
    response::Response,
};
use snafu::{ResultExt, Snafu};
use tokio::io::AsyncRead;
use tokio_util::io::ReaderStream;

#[derive(Debug, Snafu)]
/// Errors that can occur when downloading a file.
pub enum DownloadError {
    /// A failure to construct the response body.
    #[snafu(display("Failed to fetch cover."))]
    Body { source: http::Error },
}

/// Handle a request for file.
///
/// Must not be used directly from a route as that makes it vulnerable to path traversal attacks.
pub async fn handler<A: AsyncRead + Send + Unpin + 'static>(
    file_name: &str,
    reader: A,
    content_type: &str,
) -> Result<Response, DownloadError> {
    let stream = ReaderStream::new(reader);
    let body = Body::from_stream(stream);

    let response = Response::builder()
        .status(StatusCode::OK)
        .header(
            header::CONTENT_DISPOSITION,
            format!("filename=\"{file_name}\""),
        )
        .header(header::CONTENT_TYPE, content_type)
        .body(body)
        .context(BodySnafu)?;

    Ok(response)
}
