//! Handle requests for books.
use std::{io, sync::Arc};

use axum::{
    extract::{Path, State},
    http::StatusCode,
    response::Response,
};
use calibre_db::calibre::DataStoreError;
use snafu::{ResultExt, Snafu};
use tokio::fs::File;

use crate::{
    api::{
        SortOrder, TAG, download,
        error::{ErrorResponse, HttpStatus},
        paginated::{self, PaginationError},
    },
    app_state::AppState,
    data::book::{Book, Format},
    http_error,
    opds::media_type::MediaType,
};

#[derive(Debug, Snafu)]
/// Errors that can occur when retrieving books.
pub enum RetrieveError {
    /// A failure to fetch pagination data.
    #[snafu(display("Failed to fetch pagination data."))]
    Books { source: BookError },
}
impl HttpStatus for RetrieveError {
    fn status_code(&self) -> StatusCode {
        match self {
            RetrieveError::Books { source: _ } => StatusCode::INTERNAL_SERVER_ERROR,
        }
    }
}
http_error!(RetrieveError);

/// Get the first page of all books.
#[utoipa::path(
    get,
    path = "/books",
    tag = TAG,
    responses(
        (status = OK, content_type = "text/html"),
        (status = 500, description = "Server failure.", body = ErrorResponse)
    )
)]
pub async fn handler_init(State(state): State<Arc<AppState>>) -> Result<Response, RetrieveError> {
    books(&state, None, SortOrder::ASC)
        .await
        .context(BooksSnafu)
}

/// Get a paginated list of all books.
#[utoipa::path(
    get,
    path = "/books/{cursor}/{sort_order}",
    tag = TAG,
    responses(
        (status = OK, content_type = "text/html"),
        (status = 500, description = "Server failure.", body = ErrorResponse)
    )
)]
pub async fn handler(
    Path((cursor, sort_order)): Path<(String, SortOrder)>,
    State(state): State<Arc<AppState>>,
) -> Result<Response, RetrieveError> {
    books(&state, Some(&cursor), sort_order)
        .await
        .context(BooksSnafu)
}

#[derive(Debug, Snafu)]
/// Errors that can occur when fetching a book.
pub enum BookError {
    /// A failure to fetch pagination data.
    #[snafu(display("Failed to fetch pagination data."))]
    Pagination { source: PaginationError },
}

/// Render a paginated list of books.
async fn books(
    state: &Arc<AppState>,
    cursor: Option<&str>,
    sort_order: SortOrder,
) -> Result<Response, BookError> {
    paginated::render(
        "books",
        || {
            state
                .calibre
                .books(25, cursor, &sort_order.into())
                .map(|x| {
                    x.iter()
                        .filter_map(|y| {
                            Book::full_book(y, &state.calibre, &state.config.library_path)
                        })
                        .collect()
                })
        },
        |book| book.data.sort.clone(),
        |cursor| state.calibre.has_previous_books(cursor),
        |cursor| state.calibre.has_more_books(cursor),
    )
    .context(PaginationSnafu)
}

#[derive(Debug, Snafu)]
/// Errors that can occur when downloading a book.
pub enum DownloadError {
    /// A failure to fetch book data.
    #[snafu(display("Failed to fetch book data."))]
    BookData { source: DataStoreError },
    /// The requested book was not found.
    #[snafu(display("No such book."))]
    NotFound,
    /// The requested book file was not found.
    #[snafu(display("No such book."))]
    FileNotFound { source: io::Error },
    /// A failure to stream the book file.
    #[snafu(display("Failed to stream book file."))]
    Stream { source: download::DownloadError },
}
impl HttpStatus for DownloadError {
    fn status_code(&self) -> StatusCode {
        match self {
            DownloadError::BookData { source: _ } => StatusCode::INTERNAL_SERVER_ERROR,
            DownloadError::NotFound => StatusCode::NOT_FOUND,
            DownloadError::FileNotFound { source: _ } => StatusCode::NOT_FOUND,
            DownloadError::Stream { source: _ } => StatusCode::INTERNAL_SERVER_ERROR,
        }
    }
}
http_error!(DownloadError);

/// Download a book file in the specified format.
#[utoipa::path(
    get,
    path = "/book/{id}/{format}",
    tag = TAG,
    responses(
        (status = OK, content_type = "application/*"),
        (status = 500, description = "Server failure.", body = ErrorResponse)
    )
)]
pub async fn download(
    Path((id, format)): Path<(u64, String)>,
    State(state): State<Arc<AppState>>,
) -> Result<Response, DownloadError> {
    let book = state.calibre.scalar_book(id).context(BookDataSnafu)?;
    let book = Book::full_book(&book, &state.calibre, &state.config.library_path)
        .ok_or(NotFoundSnafu.build())?;
    let format = Format(format);
    let file_name = book.formats.get(&format).ok_or(NotFoundSnafu.build())?;
    let file_path = state
        .config
        .library_path
        .join(book.data.path)
        .join(file_name);
    let file = File::open(file_path).await.context(FileNotFoundSnafu)?;
    let content_type: MediaType = format.into();
    let content_type = format!("{content_type}");

    download::handler(file_name, file, &content_type)
        .await
        .context(StreamSnafu)
}
