//! Handle requests for book covers.
use std::{fs::File, io, path::Path as FilePath, sync::Arc};

use axum::{
    extract::{Path, State},
    http::StatusCode,
    response::Response,
};
use calibre_db::calibre::{Calibre, DataStoreError};
use snafu::{ResultExt, Snafu};
use tokio::fs::File as AsyncFile;

use crate::{
    api::{
        TAG,
        download::{self, DownloadError},
        error::{ErrorResponse, HttpStatus},
    },
    app_state::AppState,
    cache::{self, RetrieveThumbnailError},
    http_error,
};

#[derive(Debug, Snafu)]
/// Errors that can occur when retrieving a cover.
pub enum RetrieveError {
    /// A failure to fetch the cover.
    #[snafu(display("Failed to fetch cover."))]
    Cover { source: CoverError },
    /// A failure to open the cover file.
    #[snafu(display("Failed to open cover."))]
    CoverOpen { source: io::Error },
}
impl HttpStatus for RetrieveError {
    fn status_code(&self) -> StatusCode {
        match self {
            RetrieveError::Cover { source: _ } => StatusCode::INTERNAL_SERVER_ERROR,
            RetrieveError::CoverOpen { source: _ } => StatusCode::INTERNAL_SERVER_ERROR,
        }
    }
}
http_error!(RetrieveError);

/// Get the thumbnail for a book cover.
#[utoipa::path(
    get,
    path = "/cover/{id}/thumbnail",
    tag = TAG,
    responses(
        (status = OK, content_type = "image/jpeg"),
        (status = 500, description = "Server failure.", body = ErrorResponse)
    )
)]
pub async fn thumbnail(
    Path(id): Path<u64>,
    State(state): State<Arc<AppState>>,
) -> Result<Response, RetrieveError> {
    cover(
        &state.calibre,
        &state.config.library_path,
        &state.config.cache_path,
        id,
        |cover_path, cache_path| {
            cache::get_thumbnail(cover_path, cache_path).context(ThumbnailSnafu)
        },
    )
    .await
    .context(CoverSnafu)
}

/// Get the full-size cover for a book.
#[utoipa::path(
    get,
    path = "/cover/{id}",
    tag = TAG,
    responses(
        (status = OK, content_type = "image/jpeg"),
        (status = 500, description = "Server failure.", body = ErrorResponse)
    )
)]
pub async fn full(
    Path(id): Path<u64>,
    State(state): State<Arc<AppState>>,
) -> Result<Response, RetrieveError> {
    cover(
        &state.calibre,
        &state.config.library_path,
        &state.config.cache_path,
        id,
        |cover_path, _| File::open(cover_path).context(FileOpenSnafu),
    )
    .await
    .context(CoverSnafu)
}

#[derive(Debug, Snafu)]
/// Errors that can occur when fetching a cover.
pub enum CoverError {
    /// A failure to fetch book data.
    #[snafu(display("Failed to fetch book data."))]
    BookData { source: DataStoreError },
    /// The requested cover was not found.
    #[snafu(display("No such cover"))]
    NotFound { source: CoverFetchError },
    /// A failure to stream the cover.
    #[snafu(display("Failed to stream cover."))]
    StreamCover { source: DownloadError },
}

#[derive(Debug, Snafu)]
/// Errors that can occur when fetching a cover file.
pub enum CoverFetchError {
    /// A failure to fetch the cover thumbnail.
    #[snafu(display("Failed to fetch cover thumbnail."))]
    Thumbnail { source: RetrieveThumbnailError },
    /// A failure to open the cover file.
    #[snafu(display("Failed to open cover file."))]
    FileOpen { source: io::Error },
}

/// Generic cover handler.
async fn cover<F>(
    calibre: &Calibre,
    library_path: &FilePath,
    cache_path: &FilePath,
    id: u64,
    f: F,
) -> Result<Response, CoverError>
where
    F: Fn(&FilePath, &FilePath) -> Result<File, CoverFetchError>,
{
    let book = calibre.scalar_book(id).context(BookDataSnafu)?;
    let cover_path = library_path.join(book.path).join("cover.jpg");

    let cover = f(&cover_path, cache_path).context(NotFoundSnafu)?;
    let cover = AsyncFile::from_std(cover);
    download::handler("cover.jpg", cover, "image/jpeg")
        .await
        .context(StreamCoverSnafu)
}
