//! Handle requests for recent books.

use std::sync::Arc;

use axum::{
    extract::State,
    http::StatusCode,
    response::{Html, IntoResponse, Response},
};
use snafu::{ResultExt, Snafu};
use tera::Context;

use crate::{
    api::{
        TAG,
        books::{self, RecentBooksError},
        error::{ErrorResponse, HttpStatus},
    },
    app_state::AppState,
    http_error,
    templates::TEMPLATES,
};

/// Errors that occur during query processing.
#[derive(Debug, Snafu)]
pub enum RecentError {
    #[snafu(display("Failed to fetch recent books."))]
    RecentBooks { source: RecentBooksError },
    #[snafu(display("Failed to render template."))]
    Template { source: tera::Error },
}
impl HttpStatus for RecentError {
    fn status_code(&self) -> StatusCode {
        match self {
            RecentError::RecentBooks { source: _ } => StatusCode::INTERNAL_SERVER_ERROR,
            RecentError::Template { source: _ } => StatusCode::INTERNAL_SERVER_ERROR,
        }
    }
}
http_error!(RecentError);

/// Get recently added books
///
/// Provides a list of the 25 most recently added books.
/// The format can be either HTML or an OPDS feed, depending on the `Accept` header.
#[utoipa::path(
    get,
    path = "/recent",
    tag = TAG,
    responses(
        (status = 200, description = "List of recent books", content_type = "text/html"),
        (status = 500, description = "Error retrieving books from database", body = ErrorResponse)
    )
)]
pub async fn handler(State(state): State<Arc<AppState>>) -> Result<Response, RecentError> {
    let recent_books = books::recent(&state.calibre, &state.config.library_path)
        .await
        .context(RecentBooksSnafu)?;

    let mut context = Context::new();
    context.insert("title", "");
    context.insert("nav", "recent");
    context.insert("books", &recent_books);

    Ok(TEMPLATES
        .render("book_list", &context)
        .map(Html)
        .context(TemplateSnafu)?
        .into_response())
}
