use std::sync::Arc;

use axum::{
    extract::{Path, State},
    http::StatusCode,
    response::{Html, IntoResponse, Response},
};
use snafu::{ResultExt, Snafu};
use tera::Context;

use crate::{
    api::{
        SortOrder, TAG,
        error::{ErrorResponse, HttpStatus},
        paginated::{self, PaginationError},
        series::{self, SingleSeriesError},
    },
    app_state::AppState,
    http_error,
    templates::TEMPLATES,
};

#[derive(Debug, Snafu)]
/// Errors that can occur when retrieving series.
pub enum RetrieveError {
    /// A failure to fetch series data.
    #[snafu(display("Failed to fetch series data."))]
    Series { source: SeriesError },
}
impl HttpStatus for RetrieveError {
    fn status_code(&self) -> StatusCode {
        match self {
            RetrieveError::Series { source: _ } => StatusCode::INTERNAL_SERVER_ERROR,
        }
    }
}
http_error!(RetrieveError);

/// Get the first page of all series.
#[utoipa::path(
    get,
    path = "/series",
    tag = TAG,
    responses(
        (status = OK, content_type = "text/html"),
        (status = 500, description = "Server failure.", body = ErrorResponse)
    )
)]
pub async fn handler_init(State(state): State<Arc<AppState>>) -> Result<Response, RetrieveError> {
    series(&state, None, SortOrder::ASC)
        .await
        .context(SeriesSnafu)
}

/// Get a paginated list of all series.
#[utoipa::path(
    get,
    path = "/series/{cursor}/{sort_order}",
    tag = TAG,
    responses(
        (status = OK, content_type = "text/html"),
        (status = 500, description = "Server failure.", body = ErrorResponse)
    )
)]
pub async fn handler(
    Path((cursor, sort_order)): Path<(String, SortOrder)>,
    State(state): State<Arc<AppState>>,
) -> Result<Response, RetrieveError> {
    series(&state, Some(&cursor), sort_order)
        .await
        .context(SeriesSnafu)
}

#[derive(Debug, Snafu)]
/// Errors that can occur when fetching a series.
pub enum SeriesError {
    /// A failure to fetch pagination data.
    #[snafu(display("Failed to fetch pagination data."))]
    Pagination { source: PaginationError },
}

/// Render a paginated list of series.
async fn series(
    state: &Arc<AppState>,
    cursor: Option<&str>,
    sort_order: SortOrder,
) -> Result<Response, SeriesError> {
    paginated::render(
        "series",
        || state.calibre.series(25, cursor, &sort_order.into()),
        |series| series.sort.clone(),
        |cursor| state.calibre.has_previous_series(cursor),
        |cursor| state.calibre.has_more_series(cursor),
    )
    .context(PaginationSnafu)
}

#[derive(Debug, Snafu)]
/// Errors that can occur when retrieving a single series.
pub enum SingleError {
    /// A failure to fetch series data.
    #[snafu(display("Failed to fetch series data."))]
    Data { source: SingleSeriesError },
    /// A failure to render the template.
    #[snafu(display("Failed to render template."))]
    Render { source: tera::Error },
}
impl HttpStatus for SingleError {
    fn status_code(&self) -> StatusCode {
        match self {
            SingleError::Data { source: _ } => StatusCode::INTERNAL_SERVER_ERROR,
            SingleError::Render { source: _ } => StatusCode::INTERNAL_SERVER_ERROR,
        }
    }
}
http_error!(SingleError);

/// Get a single series and all its books.
#[utoipa::path(
    get,
    path = "/series/{id}",
    tag = TAG,
    responses(
        (status = OK, content_type = "text/html"),
        (status = 500, description = "Server failure.", body = ErrorResponse)
    )
)]
pub async fn single(
    Path(id): Path<u64>,
    State(state): State<Arc<AppState>>,
) -> Result<Response, SingleError> {
    let (series, books) = series::single(id, &state.calibre, &state.config.library_path)
        .await
        .context(DataSnafu)?;

    let mut context = Context::new();
    context.insert("title", &series.name);
    context.insert("nav", "series");
    context.insert("books", &books);

    Ok(TEMPLATES
        .render("book_list", &context)
        .context(RenderSnafu)
        .map(Html)?
        .into_response())
}
