//! Deal with cursor pagination.

use std::fmt::Debug;

use axum::response::{Html, IntoResponse, Response};
use calibre_db::calibre::DataStoreError;
use serde::Serialize;
use snafu::{ResultExt, Snafu};
use tera::Context;

use crate::templates::TEMPLATES;

#[derive(Debug, Snafu)]
/// Errors that can occur during pagination.
pub enum PaginationError {
    /// A failure to fetch pagination data.
    #[snafu(display("Failed to fetch pagination data."))]
    Fetch { source: DataStoreError },
    /// A failure to render the template.
    #[snafu(display("Failed to render template."))]
    Template { source: tera::Error },
    /// A failure to fetch previous items.
    #[snafu(display("Failed to fetch previous items."))]
    Previous { source: DataStoreError },
    /// A failure to fetch more items.
    #[snafu(display("Failed to fetch more items."))]
    More { source: DataStoreError },
}

/// Render a tera template with paginated items and generate back and forth links.
pub fn render<T: Serialize + Debug, F, S, P, M>(
    template: &str,
    fetcher: F,
    sort_field: S,
    has_previous: P,
    has_more: M,
) -> Result<Response, PaginationError>
where
    F: Fn() -> Result<Vec<T>, DataStoreError>,
    S: Fn(&T) -> String,
    P: Fn(&str) -> Result<bool, DataStoreError>,
    M: Fn(&str) -> Result<bool, DataStoreError>,
{
    let mut context = Context::new();
    context.insert("nav", template);

    let items = fetcher().context(FetchSnafu)?;
    if items.is_empty() {
        return Ok(TEMPLATES
            .render("empty", &context)
            .context(TemplateSnafu)
            .map(Html)?
            .into_response());
    }

    let first_item = items.first().unwrap();
    let last_item = items.last().unwrap();

    let (backward_cursor, forward_cursor) = (sort_field(first_item), sort_field(last_item));

    let has_previous = has_previous(&backward_cursor).context(PreviousSnafu)?;
    let has_more = has_more(&forward_cursor).context(MoreSnafu)?;

    context.insert("has_previous", &has_previous);
    context.insert("has_more", &has_more);
    context.insert("backward_cursor", &backward_cursor);
    context.insert("forward_cursor", &forward_cursor);
    context.insert(template, &items);

    Ok(TEMPLATES
        .render(template, &context)
        .context(TemplateSnafu)
        .map(Html)?
        .into_response())
}
