//! Handle caching of files, specifically book covers.

use std::{
    fmt,
    fmt::Write,
    fs::{self, File},
    io,
    path::{Path, PathBuf},
};

use sha2::{
    Digest, Sha256,
    digest::{generic_array::GenericArray, typenum::U32},
};
use snafu::{ResultExt, Snafu};
use tracing::debug;

/// Errors from converting a hash to its string representation.
#[derive(Debug, Snafu)]
pub enum HashToPathError {
    #[snafu(display("Failed to generate string representation of hash."))]
    ToString { source: fmt::Error },
}

/// Convert a hash into its path representation inside the cache directory.
///
/// First hash character is the top folder, second character the second level folder and the rest
/// is the filename.
fn hash_to_path(
    hash: GenericArray<u8, U32>,
    cache_path: &Path,
) -> Result<PathBuf, HashToPathError> {
    let mut hash_string = String::new();
    for byte in hash {
        write!(&mut hash_string, "{:02x}", byte).context(ToStringSnafu)?;
    }
    let hash = hash_string;

    let first_segment = &hash[0..1];
    let second_segment = &hash[1..2];
    let remaining_segment = &hash[2..];

    Ok(PathBuf::from(cache_path)
        .join(first_segment)
        .join(second_segment)
        .join(remaining_segment))
}

/// Errors from creating a thumbnail.
#[derive(Debug, Snafu)]
pub enum CreateThumbnailError {
    #[snafu(display("No parent for {path}."))]
    ParentDir { path: String },
    #[snafu(display("Failed to create thumbnail path {path}."))]
    ThumbnailDir { source: io::Error, path: String },
    #[snafu(display("Failed to open image {path}."))]
    ImageOpen {
        source: image::ImageError,
        path: String,
    },
    #[snafu(display("Failed to save image to path {path}."))]
    ImageSave {
        source: image::ImageError,
        path: String,
    },
}

/// Create a thumbnail for `cover_path` at `thumbnail_path`.
fn create_thumbnail(cover_path: &Path, thumbnail_path: &Path) -> Result<(), CreateThumbnailError> {
    debug!("creating thumbnail for {}", cover_path.to_string_lossy());
    let folders = thumbnail_path.parent().ok_or_else(|| {
        ParentDirSnafu {
            path: thumbnail_path.to_string_lossy(),
        }
        .build()
    })?;
    fs::create_dir_all(folders).context(ThumbnailDirSnafu {
        path: folders.to_string_lossy(),
    })?;

    const THUMBNAIL_SIZE: u32 = 512;
    let img = image::open(cover_path).context(ImageOpenSnafu {
        path: cover_path.to_string_lossy(),
    })?;
    let thumbnail = img.thumbnail(THUMBNAIL_SIZE, THUMBNAIL_SIZE);
    thumbnail
        .save_with_format(thumbnail_path, image::ImageFormat::Jpeg)
        .context(ImageSaveSnafu {
            path: thumbnail_path.to_string_lossy(),
        })?;
    debug!("saved thumbnail to {}", thumbnail_path.to_string_lossy());

    Ok(())
}

/// Errors from retrieving a thumbnail.
#[derive(Debug, Snafu)]
pub enum RetrieveThumbnailError {
    #[snafu(display("Failed to convert hash to string."))]
    HashToPath { source: HashToPathError },
    #[snafu(display("Failed to create not yet existing thumbnail."))]
    CreateThumbnail { source: CreateThumbnailError },
    #[snafu(display("Failed to open thumbnail."))]
    OpenThumbnail { source: io::Error },
}

/// Get the thumbnail for a book cover.
///
/// If a thumbnail does not yet exist, create it.
pub fn get_thumbnail(cover_path: &Path, cache_path: &Path) -> Result<File, RetrieveThumbnailError> {
    let path_str = cover_path.to_string_lossy();
    let mut hasher = Sha256::new();
    hasher.update(path_str.as_bytes());
    let hash = hasher.finalize();

    let thumbnail_path = hash_to_path(hash, cache_path).context(HashToPathSnafu)?;

    if !thumbnail_path.exists() {
        create_thumbnail(cover_path, &thumbnail_path).context(CreateThumbnailSnafu)?;
    }

    File::open(thumbnail_path).context(OpenThumbnailSnafu)
}
