//! Configuration data.

use std::{
    env, fs, io,
    net::{SocketAddr, ToSocketAddrs},
    path::{Path, PathBuf},
};

use snafu::{ResultExt, Snafu};
use tracing::info;

use crate::cli::Cli;

/// Errors from loading application configuration.
#[derive(Debug, Snafu)]
pub enum LoadError {
    /// The provided path is not a calibre library.
    #[snafu(display("{path} is not a calibre library."))]
    LibraryPath { path: String },
    /// Could not find the calibre metadata database.
    #[snafu(display("Could not find calibre metadata at {path}."))]
    MetadataPath { path: String },
    /// The listening address could not be parsed.
    #[snafu(display("Invalid listening address {listen_address}."))]
    ListeningAddressParse {
        source: io::Error,
        listen_address: String,
    },
    /// The listening address is invalid.
    #[snafu(display("Invalid listening address {listen_address}."))]
    ListeningAddress { listen_address: String },
    /// The cache directory could not be created.
    #[snafu(display("Failed to create cach directory at {path}."))]
    CacheDir { source: io::Error, path: String },
}

/// Application configuration.
#[derive(Debug, Clone)]
pub struct Config {
    /// Calibre library folder.
    pub library_path: PathBuf,
    /// Calibre metadata file path.
    pub metadata_path: PathBuf,
    /// Address to listen on.
    pub listen_address: SocketAddr,
    /// Path to data like thumbnails.
    pub cache_path: PathBuf,
}

impl Config {
    /// Check if the calibre library from `args` exists and if the calibre database can be found.
    pub fn load(args: &Cli) -> Result<Self, LoadError> {
        let library_path = Path::new(&args.library_path).to_path_buf();

        if !library_path.exists() {
            let library_path = library_path
                .as_os_str()
                .to_str()
                .unwrap_or("<failed to parse path>")
                .to_string();
            return LibraryPathSnafu { path: library_path }.fail();
        }

        let metadata_path = library_path.join("metadata.db");
        if !metadata_path.exists() {
            let metadata_path = metadata_path
                .as_os_str()
                .to_str()
                .unwrap_or("<failed to parse path>")
                .to_string();
            return MetadataPathSnafu {
                path: metadata_path,
            }
            .fail();
        }
        let listen_address = args
            .listen_address
            .to_socket_addrs()
            .context(ListeningAddressParseSnafu {
                listen_address: args.listen_address.clone(),
            })?
            .next()
            .ok_or(
                ListeningAddressSnafu {
                    listen_address: args.listen_address.clone(),
                }
                .build(),
            )?;

        let cache_path = if args.cache_path.starts_with("$TMP") {
            let cache_base = env::var("XDG_CACHE_HOME")
                .unwrap_or_else(|_| env::var("TMPDIR").unwrap_or("/tmp/".to_string()));
            PathBuf::from(&cache_base).join("little-hesinde")
        } else {
            PathBuf::from(&args.cache_path)
        };
        fs::create_dir_all(&cache_path).context(CacheDirSnafu {
            path: cache_path.to_string_lossy(),
        })?;
        info!("Using {} for cache", cache_path.to_string_lossy());

        Ok(Self {
            library_path,
            metadata_path,
            listen_address,
            cache_path,
        })
    }
}
