//! Link data.

use serde::Serialize;

use super::{media_type::MediaType, relation::Relation};
use crate::data::book::{Book, Format};

/// Link element in OPDS.
#[derive(Debug, Serialize)]
#[serde(rename = "link")]
pub struct Link {
    /// Actual hyperlink.
    #[serde(rename = "@href")]
    pub href: String,
    /// Type of the target.
    #[serde(rename = "@type")]
    pub media_type: MediaType,
    /// Relation of the target.
    #[serde(rename = "@rel")]
    pub rel: Relation,
    /// Optional link title.
    #[serde(rename = "@title")]
    #[serde(skip_serializing_if = "Option::is_none")]
    pub title: Option<String>,
    /// Optional count (how many entries at the target).
    #[serde(rename = "@thr:count")]
    #[serde(skip_serializing_if = "Option::is_none")]
    pub count: Option<u64>,
}

/// Convert a format from a book into a link where it is downloadable.
impl From<(&Book, (&Format, &str))> for Link {
    fn from(value: (&Book, (&Format, &str))) -> Self {
        let format = value.1.0.clone();
        let media_type: MediaType = format.into();
        Self {
            href: format!("/book/{}/{}", value.0.data.id, value.1.0),
            media_type,
            rel: media_type.into(),
            title: Some(value.1.0.0.clone()),
            count: None,
        }
    }
}

#[cfg(test)]
mod tests {
    use quick_xml::se::to_string;

    use super::*;

    fn init() -> Link {
        Link {
            href: "/opds".to_string(),
            media_type: MediaType::Navigation,
            rel: Relation::Start,
            title: Some("Home".to_string()),
            count: Some(10),
        }
    }

    #[test]
    fn serialize() {
        let mut link = init();
        let xml = to_string(&link).unwrap();

        assert_eq!(
            xml,
            r#"<link href="/opds" type="application/atom+xml;profile=opds-catalog;kind=navigation" rel="start" title="Home" thr:count="10"/>"#
        );

        link.title = None;
        link.count = None;
        let xml = to_string(&link).unwrap();

        assert_eq!(
            xml,
            r#"<link href="/opds" type="application/atom+xml;profile=opds-catalog;kind=navigation" rel="start"/>"#
        );
    }
}
