//! Search data.

use std::io::Cursor;

use quick_xml::{
    Reader, Writer,
    events::{BytesDecl, BytesStart, Event},
    se::to_string,
};
use serde::Serialize;
use snafu::ResultExt;

use super::error::{
    AsXmlError, BytesToUtf8Snafu, ReadXmlEventSnafu, ToStringSnafu, WriteXmlEventSnafu,
};

/// Url pointing to a location.
#[derive(Debug, Serialize)]
pub struct Url {
    /// The media type of the resource.
    #[serde(rename = "@type")]
    pub type_name: String,
    /// The URL template.
    #[serde(rename = "@template")]
    pub template: String,
}

/// Search information.
#[derive(Debug, Serialize)]
pub struct OpenSearchDescription {
    #[serde(rename = "ShortName")]
    pub short_name: String,
    #[serde(rename = "Description")]
    pub description: String,
    #[serde(rename = "InputEncoding")]
    pub input_encoding: String,
    #[serde(rename = "OutputEncoding")]
    pub output_encoding: String,
    #[serde(rename = "Url")]
    pub url: Url,
}

impl OpenSearchDescription {
    /// Serialize search information to an open search description xml.
    pub fn as_xml(&self) -> Result<String, AsXmlError> {
        let xml = to_string(&self).context(ToStringSnafu)?;
        let mut reader = Reader::from_str(&xml);
        reader.config_mut().trim_text(true);

        let declaration = BytesDecl::new("1.0", Some("UTF-8"), None);
        let mut writer = Writer::new(Cursor::new(Vec::new()));
        writer
            .write_event(Event::Decl(declaration))
            .context(WriteXmlEventSnafu)?;

        let mut search_start = BytesStart::new("OpenSearchDescription");
        search_start.push_attribute(("xmlns", "http://a9.com/-/spec/opensearch/1.1/"));

        loop {
            match reader.read_event() {
                Ok(Event::Start(e)) if e.name().as_ref() == b"feed" => writer
                    .write_event(Event::Start(search_start.clone()))
                    .context(WriteXmlEventSnafu)?,
                Ok(Event::Eof) => break,
                Ok(e) => writer.write_event(e).context(WriteXmlEventSnafu)?,
                Err(e) => return Err(e).context(ReadXmlEventSnafu)?,
            }
        }
        let result = writer.into_inner().into_inner();
        String::from_utf8(result).context(BytesToUtf8Snafu)
    }
}
